#!/usr/bin/ruby
#
# fishnet -- defines generic routines for running network code under fishnet.
#	fishnet is specialized in two files: simulator and physical

$LOAD_PATH.push('../proj')	# so we can find node.rb

# first, load all the procedures we'll need
require 'English'
require 'utility'
require 'packet'
require 'node'
require 'topology'
require 'commands'

# FishNet is an abstract class representing various ways (simulation, emulation,
# ipaq, planetlab) to execute network code.
#
# All subclasses implement:
#	fishnet.start() -- called by main when the Fishnet is ready to begin
# 	fishnet.sendPkt(to, pkt)  -- called by node software to send a packet
#		 -- to is a FishNet Address, pkt is a binary string, defined in packet.rb
#	fishnet.wakeupIn(x) -- called by node software to set a wakeup call,
#			set a timer to go off in x milliseconds, return immediately
#	fishnet.pong -- make a "pong" sound on the IPAQ, and print "PONG" in
#		simulation and emulation mode
#	fishnet.now -- return the current time in milliseconds
#		(real time on the IPAQ, simulated time on the simulator)
#
# All subclasses make upcalls to class Node; Node is implemented as a state machine,
# so all of these routines return after doing their immediate work:
#	start -- when Fishnet is ready for use
#	onReceive(from, pkt)  -- when a packet arrives, notify Node
#	onKeyboard(msg) -- when a line of text (a command) is ready from the input
#	onTimer(now) -- when a wakeup alarm triggers, now is current time (in millisec)
#

class Fishnet
	def initialize
		@numSent = 0
		@start = now
	end
	
	def address?(x) 		# check an address
		fishAddress?(x)
	end
	def packet?(x)		# check the packet size
		(1...$MaxPacketSize) === x.length
	end
	
	# in the generic class, just check the arguments
	def sendPkt(to, pkt)
		raise ArgumentError if !address?(to) or !packet?(pkt)
		@numSent += 1
	end
	
	# called by Node to schedule a timer wakeup for x milliseconds into the future
	def wakeupIn(x)
		raise ArgumentError if x < 0
	end
	
	# on the IPAQ, we redefine this to play a sound file
	def pong
		$stderr.print "PONG!!\n"
	end
	
	def stop
		$stderr.print "Fishnet exiting at time: ", 
			self.now() - @start, " msec.", 
			"\nNumber of packets sent: ", @numSent, "\n"
		exit(1)
	end
end

require 'simulator'
require 'physical'

# main -- start a fishnet
#
# First, we check ARGV to see which kind of fishnet to create:
#	-s num	[topology]	    -- simulate num nodes, connected with topology
#	-e trawler-address:trawler-port localName:localPort -- emulate, coordinating through
#					the trawler at addr/port, listen at localPort
#					localName is the local machine's hostname
#	-i fishAddress [port] 
#				-- "live" mode for the ipaq's, using fishAddress

# On Windows, you may need to set ARGV and $stdin, as follows.
# ARGV = ["-s", "3", "three.topo"]
# $fishInput = File.new("three.key")
$fishInput = nil

$fishnet = nil

case ARGV[0]
	when /-s/
		$debug = SimDebug.new(true)	# for per-node messages
		$sdebug = Debug.new(true)		# for internal simulator messages
		$fishnet = FishnetSim.new(ARGV[1].to_i, ARGV[2])
	when /-e/
		$debug = Debug.new(true)
		trawlerAddress, port = ARGV[1].split(/:/)
		localName, localPort = ARGV[2].split(/:/)		
		$fishnet = FishnetEmulator.new(
			trawlerAddress, port.to_i, localName, localPort.to_i)
	when /-i/
		$debug = Debug.new(true)
		fishAddress = ARGV[1].to_i
		localPort = ARGV[2].to_i if ARGV.length >= 3
		$fishnet = FishnetIpaq.new(fishAddress, localPort)
	else
		$stderr.print "Usage: fishnet: ",
		"-s num -- simulate with # nodes\n",
		"-e trawler-address:trawler-port localPort -- ",
			"emulate by contacting trawler at addr:port\n",
		"-i fishAddr [localPort] -- run on an ipaq, using fishAddr with optional localPort\n",
		"Note: select exactly one of -s/-e/-i\n"
		raise ArgumentError
	end
begin
	$fishnet.start		# start the simulator/emulator/ipaq execution
end
