# Packet -- defines the Fishnet packet header, used for
# 	communicating between Fishnet nodes.

# Some global constants
$BroadcastAddress = 255			# 255 signifies a broadcast
$MaxAddress = 255
$MaxPacketSize = 128				# bytes, kept small on purpose
$MaxTTL = 15					# max hop count
$MaxVCircuit = 255				# max # of virtual circuits (per node)

# Packet routing types: either normal (routed based on the destination address)
# or routed based on the virtual circuit identifier
$PktDestRoute	= 0		# route based on the destination address
$PktVCRoute 	= 1		# this packet is to traverse a virtual circuit

# test if a Fishnet address is legal, from 0..255
def fishAddress?(x)
	(0..$MaxAddress) === x
end

def virtualCircuit?(x)
	(0..$MaxVCircuit) === x
end

# Protocols we currently support
# Note that we'll add to these as the quarter progresses,
# so do not modify this file!
$PingPkt 		= 0		# packet contents is a text message
$PingReplyPkt 	        = 1
$VCReqPkt		= 2		# packet contents is an array specifying the path for this virtual circuit
$VCReplyPkt	        = 3		# packet contents contain the virtual circuit ID to use
$LinkInfoPkt	        = 4		# packet contents contain the list of neighbors for the source address
$NamePkt		= 5		# packet contents contains the machine name for the source address
$TransportPkt	= 6		# packet contents are for reliable transport

# test if a protocol is legal
def fishProtocol?(x)
	x == $PingPkt or x == $PingReplyPkt or x == $VCReqPkt or 
		x == $VCReplyPkt or x == $LinkInfoPkt or x == $NamePkt or
		x == $TransportPkt
end

# for printing error messages
def protocolToS(x)
	case x
		when $PingPkt
			"Ping"
		when $PingReplyPkt
			"Ping reply"
		when $VCReqPkt
			"Virtual circuit request"
		when $VCReplyPkt
			"Virtual circuit reply"
		when $LinkInfoPkt
			"Link state packet"
		when $NamePkt
			"Name packet"
		when $TransportPkt
			"Transport"
		else
			"Unknown protocol"
	end
end

# test if this is a legal TTL value
def ttl?(x)
	(1..$MaxTTL) === x
end

# test if this packet is legal
def packet?(type, to, from, protocol, ttl, length)

	return false if !fishProtocol?(protocol) or !ttl?(ttl) or length > $MaxPacketSize
	
	if (type == $PktDestRoute)
		fishAddress?(to) and fishAddress?(from)
	elsif (type == $PktVCRoute)
		virtualCircuit?(to) and fishAddress?(from)
	else
		false
	end
end

# Packet manages the format of packets on the wire
#	destination address/virtual circuit ID: 16 bits (depending on the routing type)
#	source address: 16 bits
#	routing type: 8 bits
#	ttl (time to live) (8 bits)
#	protocol: 8 bits
#	packet length: 8 bits
#	packet id: 32 bits
#	payload: <up to 116 bytes>


class Packet
	@@Format = "nnCCCCN"		# packet header in binary
	@@HeaderSize = 12			# number of bytes added by header
	
	# create a packet to be sent
	def initialize(type, to, from, protocol, ttl, seq, msg)
		raise ArgumentError if !packet?(type, to, from, 
				protocol, ttl, msg.length + @@HeaderSize)
		@type = type
		@dest = @vcid = to
		@src = from
		@protocol = protocol
		@msg = msg
		@ttl = ttl
		@seq = seq
	end
	
	# allow various routines to directly access the contents of the packet
	attr_reader :type, :vcid, :dest, :src, :protocol, :ttl, :seq, :msg
	attr_writer :ttl, :dest, :vcid
	
	# print the packet
	def to_s
		"Packet: " + 
			((@type == $PktVCRoute) ? ("VC: " + @vcid.to_s) : 
							(@src.to_s + "->" + @dest.to_s)) +
			" protocol: " + protocolToS(@protocol) + 
			" ttl: " + @ttl.to_s +
			" seq: " + @seq.to_s +
			" contents: " + @msg
	end
	
	# convert the packet into a binary string for sending over the wire
	# note that we increment the packet_id so that it is unique among all
	# packets sourced by this node.
	def pack
		[@dest, @src, @type, @ttl, @protocol, 
			@msg.length + @@HeaderSize, 
			@seq].pack(@@Format) + @msg
	end
	
	# This is a class method to unpack a binary string into a packet.
	#  This assumes the binary string begins with the packet header
	# Returns the new packet, or nil if the binary string is obviously corrupted.
	def Packet.unpack(str)
		dest, src, type, ttl, protocol, length, id = str.unpack(@@Format)
		msg = str[@@HeaderSize...str.length]
		return nil if !packet?(type, dest, src, protocol, ttl, str.length)
		return Packet.new(type, dest, src, protocol, ttl, id, msg)
	end

end

#This manages a link state packet, to convey a list of neighbors around the
# network for building routing tables.
#
# The format of the link state packet is very simple:
#	a variable length array of <fishnet address>
#
#  Note that this header is assumed to be the "msg" contents of a class Packet,
#  so it does not need source, destination, TTL, sequence #, etc., information.

class LinkState
	@@MaxNeighbors = 50		# so we don't exceed the packet size
	
	def initialize(neighbors)
		neighbors.each {|x| throw ArgumentError if !fishAddress?(x)}
		raise ArgumentError if neighbors.length > @@MaxNeighbors
		@neighbors = neighbors
	end
	
	attr_reader :neighbors
	
	def pack
		if @neighbors.length > 0
			@neighbors.pack("n"*@neighbors.length)
		else
			""		# return the empty string, signifying no neighbors
		end
	end
	
	def LinkState.unpack(str)
		if str.length == 0
			neighbors = []
		else
			neighbors = str.unpack("n" * (str.length/2))
			# if any neighbor is invalid, assume the packet is corrupted and discard
			neighbors.each { |x| return nil if !fishAddress?(x)}
		end
		return LinkState.new(neighbors)
	end	
end

# This manages a virtual circuit request.  Contains an array of addresses to be visited along the path.
# NOTE: the destination address in the Packet is defined to be the next to be visited, followed by
# the first element in the array below, etc.
# If VCRequest contains an empty route array, then we just use normal routing to the destination.

# The code is equivalent to the LinkState packet above; we separate them in case we want to
# modify the packet format later.
class VCRequest
	@@MaxSourceRoute = $MaxTTL
	
	def initialize(route)
		route.each {|x| throw ArgumentError if !fishAddress?(x)}
		raise ArgumentError if route.length > @@MaxSourceRoute
		@route = route
	end
	
	attr_reader :route
	
	def pack
		if @route.length > 0
			@route.pack("n"*@route.length)
		else
			""		# return the empty string, signifying no source route
		end
	end
	
	def VCRequest.unpack(str)
		if str.length == 0
			route = []
		else
			route = str.unpack("n" * (str.length/2))
			# if any part of the route is invalid, return nil
			route.each { |x| return nil if !fishAddress?(x)}
		end
		return VCRequest.new(route)
	end
end

# This conveys the reply to a virtual circuit request.
# The first field is the identifier for the upstream node to use in referring to the virtual circuit.
# The second field corresponds to the VCRequest to which this VCReply is responding.
class VCReply
	@@Format = "nN"
	@@Size = 6

	def initialize(vcID, ackNo)
		raise ArgumentError if !virtualCircuit?(vcID)
		@vcID = vcID
		@ackNo = ackNo
	end

	attr_reader :vcID, :ackNo
	attr_writer :vcID

	def pack
		[@vcID, @ackNo].pack(@@Format)
	end
	
	def VCReply.unpack(str)
		return nil unless str.length == @@Size
	        vci, ack = str.unpack(@@Format)
		return VCReply.new(vci, ack)
	end
end

# This conveys the header for reliable message transfer.
# This is carried in the payload of a Packet, and in turn the data being
# transferred is carried in the payload of the Transport packet.
#types of packets
$SYNpkt = 0		# connection setup
$ACKpkt = 1		# ack (note that the ack # comes back in the seqNo)
$FINpkt = 2		# close the connection
$DATApkt = 3		# regular data

$MaxPortNum = 255	# port numbers range from 0..255

class Transport
	@@Format = "nnnnN"
	@@Size = 12
	@@MaxPayload = $MaxPacketSize - (@@Size + 12)
	
	def initialize(srcPort, destPort, type, window, seqNo, payload)
		@srcPort = srcPort		# port to use for replies back to source
		@destPort = destPort		# pkt is delivered to this port
		@type = type			# syn, ack, fin, data
		@window = window		# advertised window (for flow control)
		@seqNo = seqNo			# sequence number in bytes
							# (for acks, the last byte received)
		@payload = payload		# data being transferred, if any
		raise ArgumentError if ((payload.length > 0 and type != $DATApkt) or
						payload.length > @@MaxPayload)
	end
	
	attr_reader :srcPort, :destPort, :type, :seqNo, :window, :payload
	
	def pack
		[@srcPort, @destPort, @type, @window, @seqNo].pack(@@Format) + @payload
	end
	
	def Transport.unpack(str)
		return nil unless str.length >= @@Size
	        srcPort, destPort, type, window, seqNo = str.unpack(@@Format)
		payload = str[@@Size...str.length]
		return nil if srcPort > $MaxPortNum or destPort > $MaxPortNum
		if type == $SYNpkt or type == $ACKpkt or type == $FINpkt
			return nil if payload.length > 0
		elsif type == $DATApkt
			return nil if payload.length > @@MaxPayload
		else
			return nil		# illegal type
		end
		return Transport.new(srcPort, destPort, type, window, seqNo, payload)
	end
end
