# Physical -- Routines for implementing fishnet on a physical network
#	most of the code is shared in common between the ipaq and 
#	the emulation environment
#
# Differences (relating to the fact that the emulator gets the topology
# 	from the trawler, and the ipaq gets the topology from broadcasting
#	to its neighbors):
#
#	an emulated node gets its fishAddress from the trawler; on the ipaq
#		the user provides the (unique) fishAddress (see back of the ipaq)
#	before sending a packet, emulator should see if there are any 
#		arp updates from trawler
#	if packet is to be broadcast, ipaq can broadcast; emulator should 
#		unicast to all neighbors in the arp cache
#	if direct send, when arp lookup fails, emulator should print an error, while
#		ipaq should broadcast
#
# For now, we ignore link characteristics (loss, delay, bw) in emulation mode;
#	we could get this info from the trawler, and enforce it here

require 'socket'
require 'rbconfig.rb'
$WINDOWS = (Config::CONFIG['host'] =~ /mswin32/)

# UDPBroadcast / UDPPort is only used by iPAQs
$UDPBroadcast = "10.0.1.255"
$UDPPort = 8888			# default local port
$MaxByteRate = 100000	# don't send more than 100KB/s

# Routines for managing an emulated or ipaq fishnet
# Note: on the IPAQ, the arp cache is not a complete list of neighbors, just
# those who have sent us a message.  So you can't use it to answer assignment 1!
#
# We keep an event queue on the emulated node and on the IPAQ, to:
#	defer sendPkt if we have sent too many packets recently, to avoid flooding the network
#	allow keyboard input to be delayed a fixed amount of time
#	to remember to call Node::onTimer

class FishnetPhysical < Fishnet
	@@Version = 3			# version num for all packets sent between machines
	@@Header = "nnnn"		# string for how we pack/unpack the header
	@@HeaderSize = 8
	@@MaxPhysPacket = $MaxPacketSize + @@HeaderSize
	
	# Constructor is called with the fishnet address of the ipaq.
	def initialize(fishAddress, localHost, localPort)
		@myAddress = fishAddress	# fishAddress must be unique
		@localPort = localPort		# this node listens to localPort
		@nextSend = 0			# when we can next send a packet
		@cmds = Commands.new	# initialize the command parser
		@events = SortedEvents.new	# list of deferred events
		@arp = Hash.new		# list of neighbors
		@sockets = []
		@socket = openbind(localHost, localPort)
		@node = Node.new(fishAddress)	# create user code
		super()
	end

	def log(*args)
		$debug.print "physical: "
		$debug.print *args
		$debug.print "\n"
	end

	def openbind(addr, port)
		log "Binding to #{addr}:#{port}."
		s = UDPSocket.open
		s.bind(addr, port)
		@sockets << s
		return s
	end
	
	attr_reader :node		# for Command:processLine
	
	# Start the node, then poll for incoming packets and keyboard commands
	def start()
		@node.start
		defer = nil

		$fishInput = nil if $fishInput = $stdin  # this is no longer valid.

		# make a list of sockets/files to select() on.
		listenon = @sockets.dup
		# select() on stdin doesn't work on windows.
		listenon << $stdin unless $WINDOWS

		loop do
			currentTime = fishTime

			if $fishInput
				if $fishInput.eof?
					$fishInput = nil
				elsif (!defer or defer <= currentTime)
					# process keyboard input
					defer = @cmds.processLine($fishInput.gets, currentTime)
					@events.push(defer) {} if defer
				end
			end

			if @events.first and @events.first.timeToOccur <= currentTime
				# process a deferred event (such as a delayed send or wakeup)
				@events.shift.upcall.call
			else
				# process all pending incoming messages.
				timeToWait = !@events.first ? nil :
					(@events.first.timeToOccur - currentTime)/1000000.0
				if $WINDOWS
					timeToWait = $MaxWait if !timeToWait or timeToWait > $MaxWait
				end
				while ready = peek(listenon, timeToWait)
					ready.each { |rin|
						if rin == $stdin
							@cmds.processLine($stdin.gets, fishTime)
						else # a socket.
							pkt, tuple = rin.recvfrom(@@MaxPhysPacket)
							inet, port, name, ipAddr = tuple
							rcvPkt(pkt, [ipAddr, port])
						end
					}
					timeToWait = 0   # next time we peek, stop if there's no data.
				end
			end
		end
	end
	
	# send a packet to its destination.  If broadcast, send it to all neighbors.
	def sendPkt(to, pkt)
		super(to, pkt)		# check that its legal
		refreshArp
		physPkt = [@@Version, to, @myAddress, 0].pack(@@Header) + pkt
		if (to == $BroadcastAddress) then
			broadcastPkt(physPkt)
		elsif (ipInfo = @arp[to]) then
			physicalSend(physPkt, ipInfo)
		else
			noArp(to, physPkt)
		end
	end
	
	# unpack an incoming UDP packet, and deliver it to the application.
	# Note that we discard any packet that arrives, but isn't for us;
	# this can occur if we start listening on a UDP port that was recently
	# used by some other node.  In the wireless case, everyone hears
	# all packets, and discards those that aren't for them.
	#
	# NOTE: since trawler tells us the arp cache contents,
	# we could add something that the trawler just told us to delete.
	#
	def rcvPkt(udpPkt, ipInfo)
		log "Received msg from #{ipInfo[0]}:#{ipInfo[1]}"
		version, dest, from, tmp = udpPkt.unpack(@@Header)
		pkt = udpPkt[@@HeaderSize..udpPkt.length]
		if version != @@Version or !fishAddress?(from)
			$stderr.print "Received odd packet ", udpPkt
		else
			@arp[from] = ipInfo	# add source address to arp cache
			if (dest == @myAddress or # and check that packet is for us
						dest == $BroadcastAddress) then
				@node.onReceive(from, pkt)
			end
		end
	end
	
	# called by node to register a wakeup alarm x milliseconds in the future
	# NOTE: fishTime reports in microseconds, so we need to normalize x
	def wakeupIn(x)
		super(x)
		@events.push(fishTime + x*1000) { 
			@node.onTimer(fishTime/1000)
			}
	end
	
	# Send a packet, but defer sending it if we have sent something else recently
	def physicalSend(pkt, ipInfo)
		log "Sending pkt to #{ipInfo[0]}:#{ipInfo[1]}"
		currentTime = fishTime
		if @nextSend <= currentTime then
			@socket.send(pkt, 0, ipInfo[0], ipInfo[1])
			@nextSend = currentTime
		else
			@events.push(@nextSend) { 
				@socket.send(pkt, 0, ipInfo[0], ipInfo[1]) 
				}
		end
		# compute when we can send the next packet, if there is one
		@nextSend += pkt.length*1000000/$MaxByteRate
	end
	
	# called by Node to get the current time, in milliseconds
	def now
		fishTime/1000
	end
end # FishnetPhysical

# specialized routines for the ipaq
class FishnetIpaq < FishnetPhysical

# use UDPPort as the local port for sending/receiving messages
# there is only one fishnet process running on each IPAQ, so ok to fix the port
def initialize(fishAddress, localPort=nil)
	$UDPPort = localPort if localPort
	super(fishAddress, "10.0.1.#{fishAddress}", $UDPPort)
	openbind("10.0.1.255", @localPort)
end

# refreshArp is only relevant to the emulator
def refreshArp()
end

def start
	@socket.setsockopt(Socket::SOL_SOCKET, Socket::SO_BROADCAST, 1)
	super
end

# On the IPAQ, this is easy!
def broadcastPkt(pkt)
	physicalSend(pkt, [$UDPBroadcast, $UDPPort])
end

# if destination for pkt is not in the ARP cache, cast a wide net!
def noArp(to, pkt)
	broadcastPkt(pkt)
end

# play a sound file.  For now, just a stub.
def pong
	super
end

end  # FishnetIpaq

# specialized routines for the emulator.  See comments at the top of this
# file for how the emulation environment differs from the IPAQ environment.
class FishnetEmulator < FishnetPhysical

# connect to the trawler, and tell them about us.
# Get back our fishnet address in return.
def initialize(trawlerName, trawlerPort, localHost, localPort)
	begin
		@trawler = TCPSocket.new(trawlerName, trawlerPort)
		throw ArgumentError if !@trawler
		@trawler.sync = true
		@trawler.print localPort, "\n"
		fishAddr = @trawler.gets
	rescue SystemCallError 
			# we get here if the trawler isn't accepting requests, or if
			# it accepted our socket, but then closed it before we can write to it
		$stderr.print "Unable to connect to the trawler at ",
				trawlerName, " ", trawlerPort
		raise
	end
	if !fishAddr or !fishAddress?(fishAddr.to_i)
		$stderr.print "Illegal fish address ", fishAddr, " returned from trawler\n"
		raise ArgumentError
	elsif fishAddr.to_i == $BroadcastAddress
		# trawler returns the broadcast address to signal
		# there's already someone using the localPort
		$stderr.print "Port ", localPort, " already in use: choose another\n"
		raise ArgumentError
	else # success!
		log "Got fishAddr from trawler #{fishAddr}"
		super(fishAddr.to_i, localHost, localPort)
	end
end

# RefreshArp -- Make sure our arp cache is up to date, by checking to see
# 	if the trawler has given us any updates
# note: if we lose our connection to the trawler, @trawler.gets will return nil
def refreshArp()
	while peek([@trawler], 0) and (str = @trawler.gets) do
		log "Got arp from trawler #{str}"
		case str
			when /\A\s*reset/
				@arp.clear
			when /\A\s*add:\s*(\d+)\s*(\d+\.\d+\.\d+\.\d+):(\d+)/
				if fishAddress?(fishAddr = $1.to_i)
					@arp[fishAddr] = [$2, $3.to_i]
				end
			when /\A\s*remove:\s*(\d+)\s*/
				@arp.delete $1
			else
				$stderr.print "Unrecognized command from trawler\n"
		end
	end
end

# if packet is to be broadcast, need to send it once directly to each neighbor
# in the ARP cache
def broadcastPkt(pkt)
	@arp.each { |fishAddr, ipInfo|  physicalSend(pkt, ipInfo) }
end

# if destination is not in the ARP cache, error message
def noArp(dest, pkt)
	$stderr.print "Not a neighbor to destination ", dest
end

end
