# Fishnet Simulator

# FishnetSim implements Fishnet in a single address space,
# by simulating message passing between the various nodes in the net.
#
# We are created by main; in turn we create k simulated "Nodes",
# running student code.  The student code running on each node sends packets, 
# receives packets, and gets keyboard
# input, just as it would if they were running on real hardware.
#
# Note that we set $currentNode before calling into the code on each node;
# this lets us keep track of which node we are executing at any given point in time.
# 
# We use a global event queue to keep track of which message to deliver
# next; messages are delivered in the same order they would be
# delivered in real time.
#
# One bit of unrealism -- code executes infinitely quickly in simulated time.
# In practice, for very fast networks, processing time can be a bottleneck
# that would need to be explicitly considered.
#
# Another bit of unrealism -- we only squash events to be delivered to a node
# if the node (or the incoming edge in the case of a packet) is failed at the 
# time when the event is generated or when it would be delivered.  
# This is (at least sometimes) incorrect in the case where the node is alive when the
# event is queued, then the node (or edge) fails,
# and then is restarted before the event would have be delivered.
# In this case we deliver it; equally reasonable would be to squash it.

# defined so we can keep track of who is invoking sendPkt
class SimNode < Node
	def initialize(x)
		@fishAddress = x
		super(x)
	end
	
	attr_reader :fishAddress
end

# Fishnet implemented as a simulator
class FishnetSim < Fishnet
	@@MaxToSimulate = $MaxAddress - 1  # without the broadcast address
	
	# Initialize the fishnet simulation
	# 	x -- number of Nodes to simulate
	# 	topoFile -- topology file to use for connecting the nodes
	def initialize(x, topoFile)
		raise ArgumentError if x <= 0 or x > @@MaxToSimulate
	
		@now = 0				# simulated time, in microseconds
		@events = SortedEvents.new	# events occuring in the future

		# create a new Node, for every simulated node, with FishnetAddress i
		@nodes = (0...x).collect { |i| SimNode.new(i)}
		
		# parse the topology file, and defer processing if we see a "time" command		
		@cmds = SimulationCmds.new	# command line parser
		@defer = @cmds.parseFile(topoFile, 0)
		@events.push(@defer) {parseRestOfFile } if @defer
		super()
	end
	
	attr_reader :nodes
	
	# process the remainder of the topology commands file
	def parseRestOfFile()
		@defer = @cmds.parseRemainder(@now)
		@events.push(@defer) { parseRestOfFile } if @defer
	end
	
	# Start each of the nodes; we're ready for nodes to start sending and
	# receiving packets.  Once we've started everybody, start processing the
	# event queue, delivering each event in turn.  First check the keyboard to
	# see if there are commands to deliver.
	def start
		$sdebug.print "Fishnet:start, topology: ", $topology.to_s, "\n"
		@nodes.each {|x| ($currentNode = x).start}
		
		@now = 1		# we don't process keyboard messages until time > 0
		keyDefer = nil
		
		# keep looping while there's work to do
		while ($fishInput and !iodone?($fishInput)) or @events.first 
			timeToWait = @events.first ? 0 : nil
			while (!keyDefer or keyDefer <= @now) and $fishInput and
					peek([$fishInput], timeToWait) and !$fishInput.eof do
				keyDefer = @cmds.processLine($fishInput.gets, @now)
				# we push a "no op" event, to ensure that we resume processing
				# the loop when time advances to equal keyDefer
				@events.push(keyDefer) {} if keyDefer
			end
			
			# in the simulation, events represent:
			#	packets to be recieved
			# 	wakeup calls
			#	deferred processing  of the topology and keyboard files
			if event = @events.shift then
				@now = event.timeToOccur
				$sdebug.print "Making an upcall at time ",
						@now, "\n"
				event.upcall.call
			end
		end
	end
	
	# called by Node to send a pkt to a peer.  Need to check the topology
	# to make sure we have a connection to the peer, and to expand the message
	# list if this is a broadcast address.
	def sendPkt(to, pkt)
		super(to, pkt)		# check arguments
		
		from = $currentNode.fishAddress
		if to == $BroadcastAddress then
			@nodes.each { |x|
				edge = $topology.liveEdge?(from, x.fishAddress)
				deliverPkt(x, from, pkt, edge) if edge
				}
		elsif edge = $topology.liveEdge?(from, to)
			deliverPkt(@nodes[to], from, pkt, edge)
		else	# should we return an error here, or fail silently?
			$stderr.print "Unable to deliver pkt: ",
				from, " -> ", to, "\n"
		end
	end
	
	# schedule a packet to be delivered to the appropriate node
	def deliverPkt(to, from, pkt, edge)
		toOccur = edge.schedulePkt(from, pkt.length, @now)
		return unless toOccur
		@events.push(toOccur) {
			# don't deliver message if link or endpoint has failed
			if $topology.liveEdge?(from, to.fishAddress) then
				$currentNode = to
				$currentNode.onReceive(from, pkt)
			end
			}
	end
	
	# called by node to register a wakeup alarm x milliseconds in the future
	def wakeupIn(x)
		super(x)		# check arguments
		toWake = $currentNode
		@events.push(@now + x*1000) {
			# this is called at a later time, when $currentNode may be redefined
			# so we have to reset currentNode; also, the node may
			# have failed in the meantime, so check that too!
			if $topology.liveNode?(toWake.fishAddress)
				$currentNode = toWake
				$currentNode.onTimer(@now/1000)
			end
			}
	end
	
	# The node has failed, and needs to be restarted
	# Just allocate a new node, and kick it.  
	# The garbage collector will reclaim the old node.
	def restartNode(x)
		@nodes[x] = SimNode.new(x)
		$currentNode = @nodes[x]
		@nodes[x].start()
	end
	
	# retrieve current time in milliseconds
	def now
		@now / 1000
	end

end  # FishnetSim

class SimDebug < Debug
	def print(*varargs)
		$stderr.print $currentNode.fishAddress, ": " if $currentNode and @debugOn
		super(varargs)
	end
end
