# trawler -- routines to manage an emulated fishnet
#
# Usage: trawler port [topology]
#	-- port is the TCP port to listen for nodes joining the fishnet
#	-- topology is a file specifying how to connect the nodes; default is all to all
#
# Note: emulated nodes use a TCP socket to talk to the trawler, but
# 	send and receive messages directly to other emulated nodes via UDP.
#	If an emulated node leaves the system, writes to its TCP socket will
#	raise an system call exception; reads will return with eof.
#  
# Works by listening for new TCP connections.  Each new connection
#	sends the UDP port it is using for listening to peers; note
# 	that multiple emulated nodes can be running on the same machine,
#	so this lets us disambiguate them.
#	The trawler replies with the fishnet address that the emulated node
#	should use, as well as the current neighbor list for that node,
#	as <fishnetAddress ipAddress:ipPort> pairs.  The trawler updates this
#	list as it changes (as emulated nodes come and go, and as the topology
#	changes, e.g., as specified in the topology file).
#
# Note that there is no disaster if the emulated node does not use the
# most up to date information about its neighbors from the trawler.  If a node
# sends to the wrong UDP port, it will be dropped if there is no one there to
# grab it, or if it is for the wrong fishnet address.
#
#  Note that the topology (from the topology file) is idealized; the emulated
# 	topology is that subset of the nodes and edges in the topology file that
#	correspond to nodes that have actually connected to the trawler.

# load all the files we'll need
require 'socket'
require 'English'
require 'utility'
require 'packet'
require 'topology'
require 'commands'

# simulate an argument list
#ARGV[0] = "8888"
#ARGV[1] = 'C:\Documents and Settings\tom\My Documents\ruby2\lib/three.topo'

# EmulatedNode -- keep track of information about an emulated node
class EmulatedNode
	# initialize internal data structures, and tell node which
	# fishAddress we've assigned it
	def initialize(tcpSock, fishAddr, ipAddr, ipPort)
		@tcpSock = tcpSock	# the tcp connection we use to talk to node
		@fishAddr = fishAddr	# the node's fishnet address
		@ipAddr = ipAddr		# the ip address of the node
		@ipPort = ipPort		# the node is listening on port #
	end
	
	attr_reader :fishAddr, :ipAddr, :ipPort
	
# The next several routines tell an emulated node about an update to its
# topology.  We check if the node is still alive before sending them the update;
# this is strictly not necessary, as the rescue will also catch the error.
# We need the rescue in any case, since the node may fail in the brief time
# between the alive? test and the actual TCP write.

	# Tell the emulated node about one of its neighbors.
	def putEdge(peer)
		@tcpSock.print "add: ", peer.fishAddr, " ",
					peer.ipAddr, ":", peer.ipPort, "\n" if alive?
	rescue SystemCallError
		$trawler.remove self
	end
	
	# tell the emulated node that one of its neighbors just left the system
	def removeEdge(peer)
		@tcpSock.print "remove: ", peer.fishAddr, "\n" if alive?
	rescue SystemCallError
		$trawler.remove self
	end
	
	# clears all neighbors for this node
	def restart
		@tcpSock.print "reset\n" if alive?
	rescue SystemCallError
		$trawler.remove self
	end
	
	# if the other side has closed the connection, peek will return true
	# and eof will be true.  Note that eof will hang until the socket is closed.
	def alive?
		return true if !peek([@tcpSock], 0)
		return !@tcpSock.eof		# don't call this unless peek is true
	end
	
	# print information about the emulated node
	def to_s
		"<TCP: " + @tcpSock.peeraddr[3] + ":" +
			@tcpSock.peeraddr[1].to_s + " Fish: " +
			@fishAddr.to_s + " UDP: " +
			@ipAddr + ":" + @ipPort + ">"
	end
end

# routines to implement the trawler
class Trawler

	def initialize
		# create the server port
		@sock = TCPServer.new(ARGV[0].to_i)
		
		# sessions is indexed by fishAddress; nil signifies no node currently
		# assigned that fishnet address
		@sessions = []
		(0..$MaxAddress).each { |x| @sessions[x] = nil}
		
	rescue SystemCallError # can get here if we aren't able to create the socket
		$stderr.print "Unable to intialize trawler socket.\n"
		raise
	end
	
	# does this port conflict with any of the other ports currently in use?
	def portConflict?(ipAddr, port)
		@sessions.each { |x|
			return true if x and x.ipAddr == ipAddr and x.ipPort == port
			}
		return false
	end

# the next several routines track changes to the topology, and update
# the emulated nodes to reflect those changes.  Note these are changes
# to the *topology*, not the set of live emulated nodes.

	# an edge was removed from the topology; notify both ends, if both are alive
	# otherwise, ignore -- they've probably found out about it anyway
	def failEdge(a, b)
		if @sessions[a] and @sessions[b] then
			@sessions[a].removeEdge(b)
			@sessions[b].removeEdge(a)
		end
	end
	
	# an edge was added to the topology
	def startEdge(a, b)
		if @sessions[a] and @sessions[b] then
			@sessions[a].putEdge(@sessions[b]) 
			@sessions[b].putEdge(@sessions[a])
		end
	end
	
	# A node was removed from the topology.  We just remove the node's edges;
	# the user has to kill the emulated node directly.
	# NOTE: this must be called *before* the change to the topology!
	def failNode(a)
		if @sessions[a]
			@sessions[a].reset
			(0..$MaxAddress).each { |i|
				if @sessions[i] and $topology.liveEdge?(i,a)
					@sessions[i].removeEdge(a)
				end
				}
		end
	end
	
	# A node rejoined the topology, so we add back in all of its edges.
	def restartNode(a)
		if @sessions[a]
			(0..$MaxAddress).each { |i|
				if @sessions[i] and $topology.liveEdge?(i,a)
					@sessions[i].putEdge(a)
				end
				}
		end
	end

# The following set of routines track changes to the set of emulated nodes.

	# check if any of the emulated nodes have quit, and if so, notify their neighbors
	def quitCheck
		$debug.print "Entering quitcheck\n"
		quitList = @sessions.collect { |x| (x and !x.alive?) ? x : nil}
		quitList.each { |x| remove x if x}
	end
	
	# An emulated node has quit, so notify its neighbors
	def remove(x)
		$debug.print "Removing node ", x.fishAddr, "\n"
		(0..$MaxAddress).each { |i| @sessions[i] = nil if @sessions[i] == x}
		@sessions.each { |i|
			if i and $topology.liveEdge?(i.fishAddr,x.fishAddr)
				i.removeEdge(x)
			end
			}
	end
	
	# We've added a new emulated node, so notify its neighbors, and vice versa
	def updateNeighbors(x)
		@sessions.each { |i|
			if i != x and i and $topology.liveEdge?(i.fishAddr, x.fishAddr)
				$debug.print "Updating neighbors ", 
						i.fishAddr, "<->", x.fishAddr, "\n"
				x.putEdge(i)
				i.putEdge(x)
			end
		}
	end
	
# Wait for a node to connect, then:
#	find out if any nodes have left while we were waiting
#		(and if so, update their neighbors so they stop sending them packets)
#	find a free fishnet Address to assign to the new node, and send it to them
#		(by writing to the node's TCP socket)
#	find out which UDP port they are listening to
#		(by reading from the node's TCP socket)
#	tell the node about all its neighbors' IP addresses and port #'s
#	tell all their neighbors with their IP address and port #
#	loop
def start
	# read in the topology file, if there is one	
	if ARGV[1] then
		@cmds = TrawlerCmds.new(false)
		@defer = @cmds.parseFile(ARGV[1], fishTime)
	else
		@cmds = TrawlerCmds.new(true)
		@defer = nil
	end
		
	while (tcpSock = @sock.accept)
		ipAddr = tcpSock.peeraddr[3]
		$debug.print "In start, got a socket: ", 
				ipAddr, ":", tcpSock.peeraddr[1], "\n"
		tcpSock.sync = true
		quitCheck		# check to see if any emulated nodes have left
		
		# finish processing the topology file
		now = fishTime
		if @defer and @defer < now
			# need to resume processing the topology file
			@defer = @cmds.parseRemainder(now)
		end
		
		begin
			x = nil
			# get the port # from the emulated node
			portStr = tcpSock.gets
			if !portStr or portStr.to_i < 1024 or 
					portConflict?(ipAddr, portStr.to_i)
				$stderr.print "Trawler: illegal port, ",
								portStr, "\n"
				tcpSock.print $BroadcastAddress, "\n"
				tcpSock.close
			else
				# find a fishAddress for the new node
				for x in (0..$MaxAddress) do
					break if !@sessions[x]
				end
		
				if x >= $MaxAddress
					$stderr.print "Trawler: out of addresses\n"
					tcpSock.print $BroadcastAddress, "\n"
					tcpSock.close
				else
					ipPort = portStr.to_i
					$debug.print "Got port ", ipPort,
						"assigning addr: ", x, "\n"
					# tell node about its fishAddress
					tcpSock.print x, "\n"
					updateNeighbors(@sessions[x] =
							EmulatedNode.new(tcpSock, 
								x, ipAddr, ipPort))
				end
			end
		rescue SystemCallError  
			# tcpSock.print can raise an exception, if the emulated node
			# crashed before we had a chance to tell it about the fishAddress
			@sessions[x] = nil if !x
			# otherwise, continue
		end
	end
end # start

end # Trawler

$debug = Debug.new(true)
$trawler = Trawler.new
$trawler.start		# run
