# Node -- file defining the data structures and code for the
# protocol stack for a node participating in the fishnet

# This code is under student control; we provide a baseline
# "hello world" example just to show how it interfaces with the FishNet classes.
#
# NOTE: per-node global variables are prohibited.
#	In simulation mode, we execute many nodes in parallel, and thus
#	they would share any globals.  Please put per-node state in class Node
#	instead.
#
# In other words, all the code defined here should be generic -- it shouldn't
# 	matter to this code whether it is executing in a simulation,
# 	in an emulation, or on the ipaq's.


# Make a ping request
class PingReq
	def initialize(to, from, msg)
		@to = to
		@from = from
		@msg = msg
	end
	
	# send the ping request
	def send
		@sent = $fishnet.now
		pkt = Packet.new(@to, @from, $PingPkt, $MaxTTL, 0, @msg) 
		$debug.print "Sending: ", pkt.to_s	
		$fishnet.sendPkt(@to, pkt.pack)
	end
	
	attr_reader :to, :msg, :sent
end

# Class Node -- the network protocol software for a single CPU.
# 	Note that this code must be written as a "state machine" --
#	each upcall must do its work and return, so that other upcalls
#	can be delivered.
class Node
	@@Wakeup = 10000		# wakeup every 10 seconds
	def initialize(fishAddr)
		@fishAddr = fishAddr	# this node's fish address
		@pinged = []		# the nodes that I've ping'ed
	end
	
	# called when fishnet is ready for this node to run.
	# All we do is start a timer to look for un-acknowledged ping's.
	def start()
		$fishnet.wakeupIn(@@Wakeup)
	end
	
	# called when a packet has arrived for this node
	# if packet is a ping, reply with the contents of the packet	
	# if packet is a ping reply, check that it matches what was sent		
	def onReceive(from, str)
		pkt = Packet.unpack(str)
		$debug.print "Receiving: ", pkt, "\n"
		if !pkt
			$stderr.print "Illegal packet ", str, "\n"
		elsif pkt.protocol == $PingPkt
			$fishnet.pong	# play the sound card when someone pings us!
			reply = Packet.new(pkt.src, @fishAddr, 
						$PingReplyPkt, $MaxTTL, 0, pkt.msg)
			$debug.print "Replying with ", reply, "\n"
			$fishnet.sendPkt(pkt.src, reply.pack)
		elsif pkt.protocol == $PingReplyPkt
			found = nil
			@pinged.each { |x|
				found = x if (pkt.src == x.to) and
					(pkt.msg == x.msg)
				}
			if found
				$debug.print "Successful ping of: ", 
					pkt.dest, ": ", pkt.msg, "\n"
				@pinged.delete found
			else
				$stderr.print "Unknown ping reply received: ",
					pkt.dest, ": ", pkt.msg, "\n"
			end
		else
			$stderr.print "Unknown protocol ", protocol, "\n"
		end
	end
	
	# called when the user types a command to this node
	# for now, send keyboard message as a ping to a neighbor	
	def onKeyboard(msg)
		if msg =~ /\A\s*(\d+):/ then
			ping = PingReq.new($1.to_i, 
						@fishAddr, $POSTMATCH)
			ping.send
			@pinged.push ping
		else
			$stderr.print "onKeyboard: Unrecognized command ", msg
		end
	end
	
	# called when Node requested a wakeup, and the time
	# is now >= the alarm time (note that the wakeup can
	# sometimes occur a bit later than requested)
	# for now, look to see if there are any unanswered pings.
	# Note that this doesn't necessarily signal an error, since the ping
	# or ping reply could have been dropped.
	def onTimer(now)
		@pinged.delete_if { |x|
			if x.sent + 10000 < now then
				$stderr.print "Message undelivered to ",
					x.to,  " after 10 seconds\n"
				true
			else
				false
			end
			}
		$fishnet.wakeupIn(@@Wakeup)
	end

end
